# Database API creation system for Zope
# Copyright (C) 2000 Shane Hathaway
# E-mail: shathaway@earthling.net or hathawsh@pi.slcc.edu
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

__version__ = '0.3'

from Globals import Persistent, HTMLFile, MessageDialog, PersistentMapping
import string, time
import OFS.SimpleItem
import Acquisition
import AccessControl
from AccessControl.PermissionRole import PermissionRole


#def _debug(msg): print msg
def _debug(msg): pass

def manage_addDatabaseAPI(self, id, title='', author='', version='', \
                          apidesc='', REQUEST=None):
    '''Adds a DatabaseAPI to a folder.'''
    self._setObject(id, DatabaseAPI(id, title, author, version, apidesc))
    return self.manage_main(self, REQUEST)

manage_addDatabaseAPIForm = HTMLFile('addDatabaseAPI', globals())


def zopeableId(s, allowed="%s%s_" % (string.letters, string.digits)):
    'Ensures an ID input by the user is valid.'
    try:
        rval = s
        for idx in range(len(rval)):
            if not rval[idx] in allowed:
                rval[idx] = '_'
        while s[0] not in string.letters:
            s = s[1:]
        return rval
    except:
        raise 'DatabaseAPI', 'Invalid ID.'

def multiSplit(s, chars):
    '''Like string.split() but permits more than one separator like
    java.util.StringTokenizer.'''
    rval = []
    start = 0
    idx = -1
    for idx in range(len(s)):
        if s[idx] in chars:
            if idx - start > 0:
                rval.append(s[start:idx])
            start = idx + 1
    idx = idx + 1
    if idx - start > 0:
        rval.append(s[start:idx])
    return rval

# The following is a number used for matching DatabaseImpl versions
# with API and Connector versions.  Anytime this version number doesn't
# match the copy kept by an object, that object knows to reconfigure.
pseudoVersion = time.time()

def _updatePseudoVersion():
    global pseudoVersion
    pseudoVersion = time.time()

def _getPseudoVersion():
    global pseudoVersion
    return pseudoVersion


class Method (Persistent):

    def __init__(self, id, args=[], rvalType='', rvalModifier='',
                 permission='', shortDesc='', desc=''):
        self.id = zopeableId(id)
        self.args = args
        self.rvalType = rvalType
        self.rvalModifier = rvalModifier
        self.permission = permission
        self.shortDesc = shortDesc
        self.desc = desc

    def __setstate__(self, state):
        # Upgrade from version 0.1.
        Persistent.__setstate__(self, state)
        if not hasattr(self, 'shortDesc'):
            self.shortDesc = ''
        if not hasattr(self, 'rvalModifier'):
            self.rvalModifier = ''

    def argsToString(self):
        '''Converts the arg list to a string with commas.
        '''
        rval = ''
        for arg in self.args:
            if len(rval) > 0:
                rval = "%s, %s" % (rval, arg)
            else:
                rval = arg
        return rval


class RecordType (Persistent):

    def __init__(self, id, fields=[], desc=''):
        self.id = zopeableId(id)
        self.fields = fields
        self.desc = desc

    def __prepFields(self):
        # Backward compatibility with v0.1:
        if len(self.fields) > 0 and not hasattr(self.fields[0], 'name'):
            f = []
            for field in self.fields:
                if not hasattr(field, 'name'):
                    f.append(Field(field))
                else:
                    f.append(field)   # How would this happen?
            self.fields = f

    def getField(self, name):
        'Gets a field by name.'
        self.__prepFields()
        for field in self.fields:
            if field.name == name:
                return field
        return None

    def listFields(self):
        'Returns the list of Field objects.'
        self.__prepFields()
        return self.fields


class Field (Persistent):

    def __init__(self, name, explanation='', reqd='', multi='',
                 domain='', tp='', note='', examples=''):
        self.name = name
        self.explanation = explanation
        self.reqd = reqd
        self.multi = multi
        self.domain = domain
        self.tp = tp
        self.note = note
        self.examples = examples


def _addSystemPermission(product, perm):
    'Adds to the permissions attached to the DatabaseAPI product.'
    if not hasattr(product, '_db_permission_list'):
        product._db_permission_list = PersistentMapping()
    product._db_permission_list[perm] = 1
    addPermission = product.aq_acquire('_manage_add_product_permission')
    try:
        addPermission(product, perm)
    except:
        # Already exists. Ignore.
        pass    

def _addPermissionSource(product, sourcePath):
    'Adds to the permission sources attached to the DatabaseAPI product.'
    if not hasattr(product, '_db_permission_sources'):
        product._db_permission_sources = PersistentMapping()
    product._db_permission_sources[sourcePath] = 1


class DatabaseAPI (Persistent,
                   OFS.SimpleItem.Item,
                   Acquisition.Implicit,
                   AccessControl.Role.RoleManager):
    '''A collection of record types and method declarations that
    form the basis for an abstract database connection.
    '''

    _isADatabaseAPI = 1
    meta_type = 'Database API'

    manage_options=(
        {'label':'Methods',            'action':'manage_methods'},
        {'label':'Record types',       'action':'manage_types'},
        {'label':'Create permissions', 'action':'manage_perms'},
        {'label':'Properties',         'action':'manage_props'},
        {'label':'Security',           'action':'manage_access'},
        {'label':'Undo',               'action':'manage_UndoForm'},
    )

    __ac_permissions__=(
        ('View management screens',        ('manage_main',
                                            'manage_methods',
                                            'manage_types',
                                            'manage_methodsEdit',
                                            'listMethods',
                                            'getMethod',
                                            'methodInfo',
                                            'manage_typesEdit',
                                            'readFields',
                                            'setFieldListVar',
                                            'listTypes',
                                            'getType',
                                            'typeInfo',
                                            'manage_props',
                                            'manage_propsEdit',)),
        ('Create Database Permissions',    ('manage_perms',
                                            'manage_permsEdit',
                                            'listPerms',)),
        ('Clean Out Database Permissions', ('listExtraPerms',
                                            'cleanExtraPerms',)),
        )

    manage_methods = HTMLFile('methods', globals())
    manage_types = HTMLFile('types', globals())
    manage_perms = HTMLFile('perms', globals())
    manage_props = HTMLFile('props', globals())
    methodInfo = HTMLFile('methodInfo', globals())
    typeInfo = HTMLFile('typeInfo', globals())
    manage_main = manage_methods

    def __init__(self, id, title, author, version, apidesc):
        self.id = id
        self.title = title
        self._methods = PersistentMapping()
        self._types = PersistentMapping()
        self._perms = PersistentMapping()
        self.author = author
        self.version = version
        self.apidesc = apidesc

    def __setstate__(self, state):
        Persistent.__setstate__(self, state)
        # Upgrade from version 0.2.
        if not hasattr(self, 'author'):
            self.author = ''
            self.version = ''
            self.apidesc = ''

    def __getRootFolder(self):
        obj = self
        while hasattr(obj, 'aq_parent') and \
              hasattr(obj.aq_parent, 'Control_Panel'):
            obj = obj.aq_parent
        return obj

    def _getProductObject(self):
        '''Gets the object that represents the DatabaseAPI product folder.'''
        try:
            obj = self.__getRootFolder()
            return obj.Control_Panel.Products.DatabaseAPI
        except:
            _debug('DatabaseAPI: cant find product folder')
            return None

    def _getSourcePath(self):
        '''Determines at least one path that will find this object.'''
        path = self.id
        try:
            obj = self.aq_parent
            while hasattr(obj, 'aq_parent') and \
                  hasattr(obj.aq_parent, 'Control_Panel'):
                path = "%s/%s" % (obj.id, path)
                obj = obj.aq_parent
            return path
        except:
            _debug('DatabaseAPI: cant find product folder')
            return None

    lastPopulated = None

    def __populatePermissions(self):
        '''Ensures all of our permissions are in the product registry.
        Performs the loading the first time and every time someone
        looks at the management interface of a different DatabaseAPI
        object.
        '''
        if DatabaseAPI.lastPopulated != self:
            DatabaseAPI.lastPopulated = self
            product = self._getProductObject()
            source = self._getSourcePath()
            _addPermissionSource(product, source)
            for perm in self._perms.values():
                _addSystemPermission(product, perm)

    def _setId(self, id):
        _updatePseudoVersion()
        self.id = id
        self.__name__ = id

    def manage_afterAdd(self, item, container):
        '''Calls __populatePermissions() right after a DatabaseAPI is added.
        Fortunately, this also gets called whenever a DatabaseAPI
        object is imported.
        '''
        self.__populatePermissions()

    def __editMethod(self, m, REQUEST):
        'Changes the properties of a Method.'
        m.args = multiSplit(REQUEST['args'], '\r\n\t ,')
        m.rvalType = REQUEST['rvalType']
        m.rvalModifier = REQUEST['rvalModifier']
        m.permission = REQUEST['permission']
        m.shortDesc = REQUEST['shortDesc']
        m.desc = REQUEST['desc']

    def manage_methodsEdit(self, REQUEST):
        'Accepts changes (add, mod, or del) to the list of Methods.'
        _updatePseudoVersion()
        if REQUEST.has_key('action.add'):
            m = Method(REQUEST['new_id'])
            self.__editMethod(m, REQUEST)
            self._methods[m.id] = m
            return self.manage_methods(self, REQUEST)
        elif REQUEST.has_key('action.mod'):
            id = REQUEST['id']
            m = self._methods[id]
            new_id = REQUEST['new_id']
            if new_id != id and len(new_id) > 0:
                if self._methods.has_key(new_id):
                    return MessageDialog(
                        title  ='ID already used',
                        message='The ID \'%s\' is already in use.' % new_id,
                        action ='manage_methods')
                else:
                    m.id = new_id
                    self._methods[new_id] = m
                    del self._methods[id]
            self.__editMethod(m, REQUEST)
            return self.manage_methods(self, REQUEST)            
        elif REQUEST.has_key('action.del'):
            selections = REQUEST['selections']  # Must be a list type.
            count = 0
            for id in selections:
                if self._methods.has_key(id):
                    del self._methods[id]
                    count = count + 1
            return MessageDialog(
                title  ='Methods deleted',
                message='%d method(s) deleted.' % count,
                action ='manage_methods')

    def listMethods(self):
        'Exposes the list of methods.'
        self.__populatePermissions()  # Take the opportunity to ensure the
        # permissions are correct.
        return self._methods.values()

    def getMethod(self, id):
        'Gets a Method object from the list of methods.'
        return self._methods[id]

    def readFields(self, REQUEST):
        'Reads the REQUEST and makes a list of fields from it.'
        newFields = []
        for key in REQUEST.keys():
            if key[:6] == 'field.':
                pos = string.find(key, '.', 6)
                if pos > 6:
                    fieldNo = string.atoi(key[6:pos])
                    subField = key[pos + 1:]
                    if fieldNo > 999 or fieldNo < 0:
                        # Ignore it.
                        continue
                    while len(newFields) < fieldNo + 1:
                        newFields.append(Field(''))
                    field = newFields[fieldNo]
                    if hasattr(field, subField):
                        setattr(field, subField, REQUEST[key])
        idx = 0
        while idx < len(newFields):
            if newFields[idx].name == '':
                del newFields[idx]
            else:
                idx = idx + 1
        return newFields

    def setFieldListVar(self, REQUEST, val):
        '''A helper method for complex DTML.  Sets up the fieldList
        variable in REQUEST.'''
        extraFields = 0
        if REQUEST.has_key('extraFields'):
            extraFields = REQUEST['extraFields']
            if type(extraFields) is type(''):
                extraFields = string.atoi(extraFields)
        elif REQUEST.has_key('doEditFields'):
            extraFields = 1
        rval = val
        REQUEST['extraFields'] = extraFields
        while extraFields > 0:
            rval.append(Field(''))
            extraFields = extraFields - 1
        REQUEST['fieldList'] = rval

    def __editType(self, t, REQUEST):
        'Changes the properties of a RecordType.'
        f = self.readFields(REQUEST)
        if len(f) > 0:
            t.fields = f
        t.desc = REQUEST['desc']

    def manage_typesEdit(self, REQUEST):
        'Accepts changes (add, mod, or del) to the list of RecordTypes.'
        _updatePseudoVersion()
        if REQUEST.has_key('action.add'):
            t = RecordType(REQUEST['new_id'])
            self.__editType(t, REQUEST)
            self._types[t.id] = t
            return self.manage_types(self, REQUEST)
        elif REQUEST.has_key('action.mod'):
            id = REQUEST['id']
            t = self._types[id]
            new_id = REQUEST['new_id']
            if new_id != id and len(new_id) > 0:
                if self._types.has_key(new_id):
                    return MessageDialog(
                        title  ='ID already used',
                        message='The ID \'%s\' is already in use.' % new_id,
                        action ='manage_types')
                else:
                    t.id = new_id
                    self._types[new_id] = t
                    del self._types[id]
            self.__editType(t, REQUEST)

            # What a hack... fixes a DTML-related bug.
            REQUEST['modifying_old'] = 0
            
            return self.manage_types(self, REQUEST)
        elif REQUEST.has_key('action.del'):
            selections = REQUEST['selections']  # Must be a list type.
            count = 0
            for id in selections:
                if self._types.has_key(id):
                    del self._types[id]
                    count = count + 1
            return MessageDialog(
                title  ='Record types deleted',
                message='%d record type(s) deleted.' % count,
                action ='manage_types')

    def listTypes(self):
        'Exposes the list of RecordTypes.'
        return self._types.values()
    
    def getType(self, id):
        'Gets a RecordType object from the list of types.'
        return self._types[id]
    
    def manage_permsEdit(self, REQUEST):
        'Accepts changes to the list of permissions.'
        perms = multiSplit(REQUEST['perms'], '\r\n')
        countAdd = 0
        countRemove = 0
        
        product = self._getProductObject()
        source = self._getSourcePath()
        _addPermissionSource(product, source)
                
        for perm in perms:
            if not self._perms.has_key(perm):
                self._perms[perm] = perm
                _addSystemPermission(product, perm)
                countAdd = countAdd + 1

        for perm in self._perms.keys():
            if not perm in perms:
                del self._perms[perm]
                countRemove = countRemove + 1
                # Don't remove system permissions.
        _updatePseudoVersion()
        return MessageDialog(
            title  ='Changes accepted',
            message='%d permission(s) added and %d permission(s) removed.' \
            % (countAdd, countRemove),
            action ='manage_perms')

    def listPerms(self):
        'Exposes the list of Permissions.'
        return self._perms.values()

    def manage_propsEdit(self, title, author, version, apidesc, REQUEST):
        'Accepts changes to certain properties.'
        self.title = str(title)
        self.author = str(author)
        self.version = str(version)
        self.apidesc = str(apidesc)
        return MessageDialog(
            title  ='Changes accepted',
            message='Property changes accepted.',
            action ='manage_props')

    def listExtraPerms(self):
        '''Lists permissions that might not be in use.
        '''
        inUse = []
        rval = []
        product = self._getProductObject()
        if hasattr(product, '_db_permission_sources'):
            # First try to find permissions that are in use
            # by other DatabaseAPI objects.  This method isn't
            # 100% accurate, but effective enough.
            sources = product._db_permission_sources
            root = self.__getRootFolder()
            for source in sources.keys():
                # Try to locate the given object and discover
                # what permissions it's providing.
                ids = string.split(source, '/')
                sourcePerms = None
                try:
                    obj = root
                    for id in ids:
                        obj = getattr(obj, id)
                    sourcePerms = obj._perms
                except:
                    pass
                if not sourcePerms:
                    # The API apparently doesn't exist anymore.
                    del sources[source]
                else:
                    # Add the permissions to the list of permissions
                    # in use.
                    for sourcePerm in sourcePerms.keys():
                        inUse.append(sourcePerm)
        # Now create the list of all permissions apparently no
        # longer in use by any DatabaseAPI.
        
        #tupl = product.aq_acquire('_getProductRegistryData')('permissions')
        #for permInfo in tupl:
        #    perm = permInfo['name']
        if hasattr(product, '_db_permission_list'):
            for perm in product._db_permission_list.keys():
                if not perm in inUse and not self._perms.has_key(perm):
                    rval.append(perm)
        return rval

    def cleanExtraPerms(self, REQUEST):
        '''Removes extra permissions specified by the user.  Does not
        remove any permissions that might be in use.'''
        product = self._getProductObject()
        removePermission = product.aq_acquire(
            '_manage_remove_product_permission')
        removeList = REQUEST['removePerms']
        _db_permission_list = None
        if hasattr(product, '_db_permission_list'):
            _db_permission_list = product._db_permission_list
        count = 0
        extraPerms = self.listExtraPerms()
        for perm in removeList:
            if perm in extraPerms:
                try:
                    removePermission(product, perm)
                except:
                    pass
                if _db_permission_list is not None:
                    try:
                        del _db_permission_list[perm]
                    except:
                        pass
                count = count + 1

        _updatePseudoVersion()
        return MessageDialog(
            title  ='Removed',
            message='%d permission(s) removed.' % count,
            action ='manage_perms')
