# Database API creation system for Zope
# Copyright (C) 2000 Shane Hathaway
# E-mail: shathaway@earthling.net or hathawsh@pi.slcc.edu
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
'''The implementation of a DatabaseAPI object.'''

__version__ = '0.4'

import Globals
from Globals import Persistent, HTMLFile, MessageDialog, PersistentMapping
import string, thread, time, urllib
from OFS.Folder import Folder
from OFS.DTMLMethod import DTMLMethod
from OFS.CopySupport import CopyError, eNotSupported
from AccessControl.PermissionRole import PermissionRole
try:
    from AccessControl.User import verify_watermark
except:
    # Pre-Zope 2.1.5.
    def verify_watermark(user): pass
import Acquisition
from DatabaseAPI import _getPseudoVersion, _updatePseudoVersion

#def _debug(msg): print msg
def _debug(msg): pass


_debuggingNow = None

def debug_aq(obj):
    '''Prints the acqusition ancestry of an object.
    For debugging only.
    '''
    global _debuggingNow
    if _debuggingNow is not None:
        return   # Already printing a trace.
    _debuggingNow = obj
    print 'Acquisition trace:'
    while 1:
        print obj
        if hasattr(obj, 'aq_parent'):
            obj = obj.aq_parent
        else:
            break
    _debuggingNow = None


def manage_addDatabaseConnector(self, id, title='', api_id='', REQUEST=None):
    '''Adds a DatabaseAPI to a folder.'''
    self._setObject(id, DatabaseConnector(id, title, api_id))
    return self.manage_main(self, REQUEST)

manage_addDatabaseConnectorForm = HTMLFile('addDatabaseConnector', globals())

def _listAvailablePeers(self, includeParent, attributeToFind):
    'Finds objects with a given attribute in self or parent.'
    rval = []
    lst = []
    if includeParent:
        if hasattr(self, 'objectValues'):
            lst = list(self.objectValues())
        parent = locateTrueParent(self)
    else:
        # Adding a new DatabaseConnector.
        parent = self
    if hasattr(parent, 'objectValues'):
        lst = lst + list(parent.objectValues())
    for obj in lst:
        if hasattr(obj, attributeToFind) and getattr(obj, attributeToFind):
            id = obj.id
            if id not in rval:
                rval.append(id)
    return rval

def listAvailableAPIs(self, includeParent):
    '''Looks only in connector and parent for DatabaseAPI objects
    and returns a list of them.'''
    return _listAvailablePeers(self, includeParent, '_isADatabaseAPI')


class APIAndImpl:
    '''__listAllMethods returns a list of objects of this type.
    '''

    def __init__(self, id, api, impl, acquiredOrPython=0):
        self.id = id
        self.api = api
        self.impl = impl
        self.acquiredOrPython = acquiredOrPython
        self.protected = impl is not None and \
                         hasattr(impl, '_dbc_protected') and \
                         impl._dbc_protected

    def _copyAttrs(self):
        '''Gathers info from the API and implementation.
        '''
        self.__copyAttr(self.impl, 'meta_type', 'Method')
        self.__copyAttr(self.impl, 'icon')
        self.__copyAttr(self.impl, 'title')
        self.__copyAttr(self.api, 'shortDesc', '')

    def __copyAttr(self, src, name, defaultVal=None):
        '''Copies an attribute from src to self, if available.
        '''
        if src and hasattr(src, name):
            setattr(self, name, getattr(src, name))
        else:
            setattr(self, name, defaultVal)

    def hasAPI(self):
        return self.api is not None

    def hasImpl(self):
        return self.impl is not None

    def isAnAPI(self):
        return hasattr(self.impl, '_isADatabaseAPI') \
               and self.impl._isADatabaseAPI

    def isADatabase(self):
        return hasattr(self.impl, '_isADatabaseConnector') \
               and self.impl._isADatabaseConnector

    def isTestOk(self):
        '''Returns true only if has callable impl and has no API
        or the arg count is zero.
        '''
        if not self.isAnAPI() and not self.isADatabase():
            if self.hasImpl() and callable(self.impl):
                if not self.protected:
                    return self.api is None or len(self.api.args) < 1
        return 0


def locateTrueParent(obj):
    '''This is necessary to make it possible to locate peer objects
    without being fooled by unusual request paths.  Finds
    the topmost parent that contains the specified object.'''
    id = str(obj.aq_base.id)
    try:
        # Sanity check.
        if not hasattr(obj.aq_parent, id):
            return None
    except:
        return None
    myBase = obj.aq_base
    obj = obj.aq_parent
    # First skip over objects of the same name but which
    # are not the same object.
    while 1:
        testObj = getattr(obj, id)
        if hasattr(testObj, 'aq_base') and testObj.aq_base is myBase:
            # Either in an attribute or through acquisition,
            # the object has been located.
            break
        elif hasattr(obj, 'aq_parent'):
            # Keep searching.
            obj = obj.aq_parent
        else:
            # Object not found?
            return None
    # Now find the innermost parent that contains this object.
    while hasattr(obj, 'aq_parent') and \
          hasattr(obj.aq_parent, id) and \
          getattr(obj.aq_parent, id).aq_base is myBase:
        obj = obj.aq_parent
    return obj


_list_type = type([])
_tuple_type = type(())


class DatabaseConnector (Folder):
    '''A collection of methods that implements a DatabaseAPI.
    '''

    meta_type = 'Database Connector'
    _isADatabaseConnector = 1

    manage_options=(
        {'label':'Methods',             'action':'manage_implMethods'},
        {'label':'Modifiers',           'action':'manage_implModifiers'},
        {'label':'Properties',          'action':'manage_implProps'},
        {'label':'API',                 'action':'manage_apiAccess'},
        {'label':'Import/Export',       'action':'manage_importExportForm'},
        {'label':'Security',            'action':'manage_access'},
        {'label':'Undo',                'action':'manage_UndoForm'},
        {'label':'Find',                'action':'manage_findFrame',
         'target':'manage_main'},
        )

    __ac_permissions__=(
        ('Add Documents, Images, and Files', ('manage_addDTMLMethod2',)),
        ('View management screens',          ('manage_implMethods',
                                              'manage_main',
                                              'manage_implModifiers',
                                              'manage_implProps',
                                              'manage_apiAccess',
                                              'manage_showAddForm',
                                              'manage_addMethodForm',
                                              'manage_makeCopy',
                                              'manage_propsEdit',
                                              'manage_test',
                                              'manage_modifiers',
                                              'getAPIMethods',
                                              'getAPITypes',
                                              'listAvailableAPIs2',
                                              'getAPIObject',
                                              'usesAPI',
                                              )),
        ('Use Database Methods',             ('db',),
                                               ('Anonymous', 'Manager',)),
        )

    manage_implMethods = HTMLFile('implMethods', globals())
    manage_implModifiers = HTMLFile('implModifiers', globals())
    manage_implProps = HTMLFile('implProps', globals())
    manage_main = manage_implMethods

    # Change some of the default security settings so that the
    # database methods are protected.  These settings can be overridden
    # easily using the management interface.
    # Tuple = Don't acquire permissions settings. List = acquire.
    _View_Permission = ('Manager', 'Owner',)
    _Access_contents_information_Permission = ('Manager', 'Owner',)

    def __init__(self, id, title, api_id):
        self.id = id
        self.title = title
        self.api_id = api_id
        self._v_createLock = thread.allocate_lock()

    def __setstate__(self, state):
        Persistent.__setstate__(self, state)
        # Prepare a _v_createLock attribute.
        self._v_createLock = thread.allocate_lock()

    # The following just update the pseudoVersion as a result of
    # some important change that may affect the database object.
    def _setId(self, id):
        _updatePseudoVersion()
        self.id = id
        self.__name__ = id

    def _setOb(self, id, object):
        _updatePseudoVersion()
        setattr(self, id, object)

    def _delOb(self, id):
        _updatePseudoVersion()
        delattr(self, id)

    def getAPIObject(self):
        'Finds the DatabaseAPI object.'
        if len(self.api_id) < 1:
            raise 'DatabaseConnectorException', 'API object not found.'
        # Get from within our own folder if possible.  This is the
        # most reliable source.
        if hasattr(self.aq_base, self.api_id):
            obj = getattr(self, self.api_id)
            if hasattr(obj, '_isADatabaseAPI') and obj._isADatabaseAPI:
                return obj            
        # Look in parent.  If we looked at the parents of parent, it might
        # be possible to trick us into using the wrong API object, or
        # if a sysadmin wanted to keep an API a secret, DatabaseConnector
        # could be used to expose it.
        parent = locateTrueParent(self)
        if hasattr(parent, 'aq_base') and \
           hasattr(parent.aq_base, self.api_id):
            obj = getattr(parent, self.api_id)
            if hasattr(obj, '_isADatabaseAPI') and obj._isADatabaseAPI:
                return obj
        raise 'DatabaseConnectorException', 'API object not found.'

    def listAvailableAPIs2(self):
        return listAvailableAPIs(self, 1)

    def getAPIMethods(self):
        'Returns the list of methods declared in the API.'
        try:
            return self.getAPIObject()._methods
        except 'DatabaseConnectorException':
            return {}

    def getAPITypes(self):
        'Returns the list of record types declared in the API.'
        try:
            return self.getAPIObject()._types
        except 'DatabaseConnectorException':
            return {}

    def isAPIReady(self):
        'Tries to locate the API object.'
        try:
            self.getAPIObject()
            return 1
        except 'DatabaseConnectorException':
            return 0
        
    def usesAPI(self):
        '''Determines whether this database uses an API or not.
        '''
        return len(self.api_id) > 0
        
    def manage_listMethods(self):
        '''Lists all methods from the API or implementation.'''
        list = self.__listAllMethods()
        for method in list:
            method._copyAttrs()  # Makes attribute access easier in DTML.
        return list

    def __listAllMethods(self):
        '''Lists all methods from the API, mixing in implementation
        info, methods not in the API, and subclassed Python methods.
        '''
        apiMethods = self.getAPIMethods()
        rval = []
        objectIds = self.objectIds()
        for id in objectIds:
            ob = self._getOb(id)
            if apiMethods.has_key(id):
                apiMethod = apiMethods[id]
                rval.append(APIAndImpl(id, apiMethod, ob))
            else:
                rval.append(APIAndImpl(id, None, ob))
        for id in apiMethods.keys():
            if id not in objectIds:
                apiMethod = apiMethods[id]
                if hasattr(self.aq_base, id):
                    # An implementation was found in a subclass.
                    rval.append(APIAndImpl(id, apiMethod,
                                           getattr(self.aq_base, id), 1))
                else:
                    rval.append(APIAndImpl(id, apiMethod, None))
        return rval

    def manage_getAPIMethod(self, id):
        '''Returns the API declaration for a method.
        '''
        apiMethods = self.getAPIMethods()
        if apiMethods.has_key(id):
            return apiMethods[id]
        return None

    def filter_impl_types(self, user, implTypes):
        '''Returns a list of method implementation types
        filtered through the permissions mechanisms and
        converted to a list of dictionaries containing
        name and filename information.'''
        metas = self.filtered_meta_types(user)
        # metas is now a tuple of dictionaries
        # containing name and action attrs.
        rval = []
        for implType in implTypes:
            name, filename = implType
            if name[0] == '/':
                # Always list.
                rval.append({'name':name[1:], 'filename':filename})
            else:
                for meta in metas:
                    if meta['name'] == name:
                        rval.append({'name':name, 'filename':filename})
                        break
        return rval
        
    def manage_showAddForm(self, REQUEST, filename=None):
        'Shows an add form of the given name.'
        if filename is None:
            filename = REQUEST['filename']
        filename = string.replace(filename, '.', '_')
        if len(filename) < 1:
            return self.manage_implMethods(self, REQUEST)
        file = HTMLFile(filename, globals())
        file = file.__of__(self)   # Set up acquisition in the DTML file.
        return file(self, REQUEST)
        
    def manage_addMethodForm(self, REQUEST):
        'Searches the REQUEST fields to find which add form to show.'
        toFind = 'methodChoice_'
        toFindLen = len(toFind)
        for item in REQUEST.items():
            try:
                key, value = item
                if value and len(value) > 0 and key and len(key) >= toFindLen \
                   and key[:toFindLen] == toFind:
                    REQUEST['new_id'] = key[toFindLen:]
                    return self.manage_showAddForm(REQUEST, value)
            except:
                # There is no __len__ attribute on some REQUEST field.
                pass
        self.manage_implMethods(self, REQUEST)

    default_dm_html2=''

    def manage_addDTMLMethod2(self, id, title='', file='', REQUEST=None, submit=None):
        """Add a DTML Method object with the contents of file. If
        'file' is empty, default document text is used.
        The only difference (currently) between this method
        and addDTMLMethod() is we have customized default text.
        """
        if type(file) is not type(''): file=file.read()
        if not file: file=self.default_dm_html2
        ob=DTMLMethod(file, __name__=id)
        ob.title=title
        id=self._setObject(id, ob)
        if REQUEST is not None:
            try: u=self.DestinationURL()
            except: u=REQUEST['URL1']
            if submit==" Add and Edit ": u="%s/%s" % (u, urllib.quote(id))
            REQUEST.RESPONSE.redirect(u+'/manage_main')
        return ''

    def manage_propsEdit(self, title, api_id, REQUEST=None):
        '''Accepts changes to the properties.'''
        self.title = title
        if api_id != self.api_id:
            self.api_id = api_id
            try:
                self.getAPIObject()
                msg = 'The database API \'%s\' has been located.' % api_id
            except 'DatabaseConnectorException':
                msg = 'Unable to locate the database API \'%s\'.' % api_id
        else:
            msg = 'Changes accepted.'
        _updatePseudoVersion()
        return MessageDialog(
            title  = 'Changes accepted.',
            message = msg,
            action = 'manage_implProps')

    def manage_apiAccess(self, REQUEST):
        '''Opens the API object if available.
        '''
        try:
            api = self.getAPIObject()
            REQUEST.RESPONSE.redirect(api.absolute_url() + '/manage_workspace')
        except 'DatabaseConnectorException':
            return MessageDialog(
                title  = 'API not found.',
                message = 'Unable to open the API because ' \
                'it has not been located.',
                action = 'manage_implProps')

    def manage_makeCopy(self, src_id, dest_id, REQUEST):
        '''Makes a copy of a specified object and places it as dest_id.
        See the DTML file add/copy.dtml for details.
        '''
        ob = getattr(self, src_id)
        if not ob.cb_isCopyable():
            raise CopyError, eNotSupported % absattr(ob.id)
        ob = ob._getCopy(self)
        ob.manage_afterClone(ob)
        ob._setId(dest_id)
        self._setObject(dest_id, ob)
        return self.manage_implMethods(self, REQUEST)

    def db(self):
        '''Returns an instance of the database attached
        to self.  Permissions will be inherited from self.
        '''
        return self._getIsolatedDatabase()._openInContextOf(self)

    def manage_test(self, methodName, REQUEST):
        '''Gets an instance of the database object and calls the
        specified method on it.'''
        db = self.db()
        if hasattr(db, methodName):
            m = getattr(db, methodName)
            if REQUEST.has_key('AUTHENTICATED_USER'):
                user = REQUEST['AUTHENTICATED_USER']
                verify_watermark(user)
                roles = m.__roles__
                if user.has_role(roles):
                    return m()
            raise 'ValidationError', \
                  'You do not have the required database permission.'
        else:
            return MessageDialog(
                title  = 'Method not found.',
                message = 'Method not found or is not callable: %s' \
                % methodName,
                action = 'manage_implMethods')

    def manage_modifiers(self, REQUEST):
        '''Changes the modifiers for each method.  Currently
        the only modifier is the "protected" modifier.
        '''
        for obj in self.objectValues():
            fieldName = '%s.protected' % obj.id
            if REQUEST.has_key(fieldName):
                prot = int(REQUEST[fieldName])
                origProt = 0
                try: origProt = obj._dbc_protected
                except AttributeError: pass
                if origProt != prot:
                    # Change the protected status.
                    if prot:
                        obj._dbc_protected = 1
                    else:
                        try: delattr(obj, '_dbc_protected')
                        except: pass
        _updatePseudoVersion()
        return self.manage_implMethods(self, REQUEST)

    def _getIsolatedDatabase(self):
        '''Returns the database object.  We may need to create it.
        We wait until now to create it because now we can use
        acquisition, whereas in __init__ and __setstate__ we cannot.
        '''
        try:
            if self._v_pseudoVersion == _getPseudoVersion():
                return self._v_isolDb
        except AttributeError:
            pass   # Not cached.

        lock = self._v_createLock
        try:
            if lock.acquire(0):  # The first thread starts.
                isolDb = self._makeIsolatedDatabase()
                self._v_isolDb = isolDb
                self._v_pseudoVersion = _getPseudoVersion()
            else:
                lock.acquire(1) # Other threads wait.
                isolDb = self._v_isolDb
        finally:
            lock.release()
        return isolDb

    def _makeIsolatedDatabase(self):
        '''
        Creates an object that represents the DatabaseConnector object
        with permissions set up and all methods linked to the true
        implementations.
        '''
        # See App/class_init.py (in a Zope 2.1 installation)
        # for details on the method to set up
        # permissions and roles in an arbitrary object.
        _debug('Creating a new isolated database object.')
        db = FullDatabaseImpl()

        # Avoid creating multiple PermissionRole instances that
        # refer to the same permission by caching them in permRoles.
        permRoles = {}

        methods = self.__listAllMethods()
        for method in methods:
            self.__createAttr(db, method, permRoles)
        return db

    def __createAttr(self, db, method, permRoles):
        '''Adds an attribute to a DatabaseImpl.
        '''
        impl = method.impl
        isSubDb = 0
        if impl is None:
            # No implementation.
            if method.hasAPI() and method.api.rvalModifier == 'multiple':
                attr = Forwarder(blankMultipleFunction)
            else:
                attr = Forwarder(blankSingleFunction)
        elif hasattr(impl, '_isADatabaseConnector') and \
             impl._isADatabaseConnector:
            # Provide direct access to the sub-database,
            # but don't use the cached version of the sub-database
            # since it may be confusing for the cached version to
            # have some of the properties set below.
            attr = impl._makeIsolatedDatabase()
            isSubDb = 1
        else:
            # It's a method implementation.
            if hasattr(impl, 'aq_base'):
                # Remove all acquisition wrappers first.
                impl = impl.aq_base
            attr = self.__wrapMethod(impl, db, method)

        # Setup permissions on the attribute.
        if method.hasAPI():
            perm = method.api.permission
        else:
            perm = None
        permRole = None

        name = method.id
        if not perm:
            if name == 'manage' or name[:7] == 'manage_':
                # Create default permissions.
                permRole = ('Manager',)
        else:
            # Create a PermissionRole object from the named permission.
            # Using acquisition, PermissionRole will automatically
            # determine the roles that have the required permission.
            if not permRoles.has_key(perm):
                permRole = PermissionRole(perm)
                permRoles[perm] = permRole
            else:
                permRole = permRoles[perm]
        if permRole:
            attr.__roles__ = permRole

        # Add a little more information to the attribute.
        if hasattr(attr, '_need__name__') and attr._need__name__:
            setattr(attr, '__name__', name)
        if method.hasAPI() and method.api.shortDesc:
            setattr(attr, '__doc__', method.api.shortDesc)
        # Finally, put the attribute into the DatabaseImpl object.
        db._addAttr(method.id, attr, isSubDb, method.protected)

    def __wrapMethod(self, impl, db, method):
        '''Puts the method inside a MethodTrampoline.
        '''
        apiObj = None
        rvalType = None
        if method.hasAPI():
            apiObj = method.api
            if method.api.rvalType:
                types = self.getAPITypes()
                if types and types.has_key(method.api.rvalType):
                    rvalType = types[method.api.rvalType]
        # We use a trampoline so that there's no need to
        # make changes to the original methods when setting
        # up permissions, and to hide other methods available
        # in the callable object.
        return MethodTrampoline(impl, method.id, apiObj, rvalType)


def blankSingleFunction(*args, **kw): pass

def blankMultipleFunction(*args, **kw): return []


class GenericDatabaseRecord: pass


class MethodTrampoline (Acquisition.Implicit):
    '''A method trampoline calls the specified method, filtering
    arguments (for security and portability) and returned values.
    '''
    __roles__ = Acquisition.Acquired
    #_isAMethodTrampoline = 1
    
    def __init__(self, toCall, id, api, rvalType):
        self._v_toCall = toCall
        self._v_id = id

        argTypes = []
        multiple = 0
        self._v_freeArgs = (api is None)
        if api != None:
            argTypes = api.args
            multiple = (api.rvalModifier == 'multiple')

        self._v_argTypes = argTypes
        self._v_multiple = multiple
        self._v_rvalType = rvalType
        if rvalType and rvalType.fields:
            self._v_firstRvalField = rvalType.fields[0].name
        else:
            self._v_firstRvalField = None
        self._v_usesKeywordArgs = 1
        try:
            # Some method types, such as Python methods,
            # require numbered arguments, while others, such as
            # DTML methods, require keyword arguments.
            if toCall.meta_type == 'Python Method':
                self._v_usesKeywordArgs = 0
        except:
            pass

    def this(self):
        # Needed by PythonMethod.
        return self.aq_parent

    def __getattr__(self, name):
        '''Provides access to attributes of the callable object.
        '''
        toCall = self._v_toCall
        # First check whether toCall has the attribute;
        # if it does not then aq_parent (that's us) will
        # be searched, resulting in infinite recursion.
        if hasattr(toCall.aq_base, name):
            return getattr(toCall, name)
        else:
            raise AttributeError, name

    def __call__(self, *args, **kw):
        '''Executes the method assigned to this object.
        '''
        #print 'calling', self._v_id
        if self._v_freeArgs:
            # Pass arguments through unchanged.
            newArgs = args
            newKw = kw
        else:
            newArgs = []
            newKw = {}
            argc = len(args)
            argTypes = self._v_argTypes
            if self._v_usesKeywordArgs:
                # Translates every argument into a named argument
                # in order to cleanly call the method.  Uses the list of
                # arguments from the DatabaseAPI to do this.
                # Creates only keyword arguments.
                if argc > 0:
                    if argc > len(argTypes):
                        raise 'TypeError', ('Too many arguments to %s() call:'\
                                            ' %d expected, got %d' % \
                                            (self._v_id, len(argTypes), argc))
                    for idx in range(argc):
                        newKw[argTypes[idx]] = args[idx]
                if len(kw) > 0:
                    for name, val in kw.items():
                        if name in argTypes:
                            newKw[name] = val
                newKw['REQUEST'] = self.aq_parent.REQUEST
                newKw['client'] = self.aq_parent
            else:
                # Creates only numbered arguments.
                idx = 0
                argTypesC = len(argTypes)
                while idx < argTypesC:
                    argName = argTypes[idx]
                    if argName == 'REQUEST':
                        val = self.aq_parent.REQUEST
                    elif argName == 'client':
                        val = self.aq_parent
                    elif idx < argc:
                        val = args[idx]
                    elif kw.has_key(argName):
                        val = kw[argName]
                    else:
                        raise 'TypeError', ('Argument not supplied to %s():' \
                                            ' %s.' % (self._v_id, argName))
                    idx = idx + 1
                    newArgs.append(val)

        # Call the method.
        rval = apply(self._v_toCall, newArgs, newKw)

        # Convert the returned value to the specified type.
        if self._v_multiple:
            if rval is None:
                rval = []
            elif hasattr(rval, '__getitem__'):
                # List compatible.
                rval = list(rval)
            else:
                typ = type(rval)
                if typ is not _list_type and typ is not _tuple_type:
                    # Assume only one row was returned.
                    rval = [rval]
        if self._v_rvalType:
            if self._v_multiple:
                # Check whether the first record needs to be
                # converted.  If not, assume the rest don't either.
                if len(rval) > 0 and self._needConvert(rval[0]):
                    for idx in range(len(rval)):
                        if self._needConvert(rval[idx]):
                            rval[idx] = self._convert(rval[idx])
            else:
                if self._needConvert(rval):
                    rval = self._convert(rval)
                
        # Return it.
        #print 'called', self._v_id, 'rval =', rval
        return rval

    def _needConvert(self, obj):
        frf = self._v_firstRvalField
        return frf and not hasattr(obj, frf)

    def _convert(self, obj):
        '''Converts the input object to the required type if
        possible.
        '''
        fields = self._v_rvalType.fields
        if hasattr(obj, 'has_key') and hasattr(obj, 'get'):
            # Dictionary compatible.
            rval = GenericDatabaseRecord()
            for field in fields:
                fname = field.name
                if obj.has_key(fname):
                    val = obj.get(fname)
                    setattr(rval, fname, val)
                else:
                    setattr(rval, fname, None)
        else:
            typ = type(obj)
            if typ == _list_type or typ == _tuple_type:
                # List compatible.
                rval = GenericDatabaseRecord()
                inputLen = len(obj)
                for idx in range(len(fields)):
                    if idx < inputLen:
                        val = obj[idx]
                    else:
                        val = None
                    setattr(rval, fields[idx].name, val)
            else:
                print obj
                raise 'DatabaseError', ( \
                    'Method %s returned an object ' \
                    'that cannot be converted to \'%s\': %s' \
                    % (self._v_id, self._v_rvalType.id, obj))
        return rval


class FullDatabaseImpl (Acquisition.Implicit):
    '''Database objects returned by _makeIsolatedDatabase() will use this
    class as a base.
    '''
    __roles__ = Acquisition.Acquired
    REQUEST = Acquisition.Acquired

    def __init__(self):
        self.__pubAttrs = {}

    def _addAttr(self, id, a, isSubDb, isProtected):
        setattr(self, id, a)
        if not isProtected:
            self.__pubAttrs[id] = (a, isSubDb)

    def _getPublicTuple(self, name):
        '''Retrieves information about public attributes only.
        '''
        if self.__pubAttrs.has_key(name):
            return self.__pubAttrs[name]
        raise AttributeError, name

    def _openInContextOf(self, context):
        '''Opens two branches of acquisition at once: the
        public branch and the full branch.
        '''
        fullDb = self.__of__(context)
        pubDb = PublicDatabaseImpl(fullDb).__of__(context)
        return pubDb


class PublicDatabaseImpl (Acquisition.Implicit):
    '''Provides access to public attrs but not protected attrs.
    Switches to the "full" branch upon call of any public method.
    This, along with MethodTrampoline, may be a candidate for
    implementation in C++.
    '''
    REQUEST = Acquisition.Acquired

    def __init__(self, fullDb):
        self.__fullDb = [fullDb]
        self.__roles__ = fullDb.aq_base.__roles__

    def __getattr__(self, name):
        '''When accessing a sub-DatabaseConnector, returns a public
        version of it, otherwise the attr acquiring from self.
        '''
        # Get the value of fullDb without being affected
        # by the implicit wrapping.
        fullDb = self.__fullDb[0]
        attr, isSubDb = fullDb._getPublicTuple(name)
        if isSubDb:
            # Set up acquisition for both branches: the full database
            # and the public database.
            subDb = attr.__of__(fullDb)
            pubDb = PublicDatabaseImpl(subDb).__of__(self)
            return pubDb
        elif callable(attr):
            # Access a method which, upon calling, will access
            # the full database branch rather than the public database.
            trueAttr = attr.__of__(fullDb)
            f = Forwarder(trueAttr).__of__(self)
            return f
        else:
            # A normal attribute.
            return attr.__of__(self)


class Forwarder (Acquisition.Explicit):
    '''A Forwarder simply calls a method while providing no access to
    other attrs that might be present in the method.  Also provides a
    place to put attrs that should not be associated with the method itself.
    '''
    # Don't ever try to provide access to other attrs in the method,
    # BTW, because calls to the other attrs will not switch
    # to the "protected" context... unless, of course, you're
    # willing to write a bunch of code.
    def __init__(self, attr):
        # Don't let attr be affected by implicit wrapping:
        #   put it in a list.  Acq machinery won't see it.
        self._attr = [attr]
        try:
            self.__roles__ = attr.aq_base.__roles__
        except:
            self.__roles__ = Acquisition.Acquired

    def __call__(self, *args, **kw):
        attr = self._attr[0]
        return apply(attr, args, kw)


Globals.default__class_init__(DatabaseConnector)

